import os
import datetime
import logging
import json
import re
from kubernetes import client, config, utils
from uuid import uuid1

# Note: apis readme: kubernetes-client/python/kubernetes/docs/CoreV1Api.md
def get_current_context():
    return config.list_kube_config_contexts()[1]

class Api:
    """Class to interact with some K8s resources"""


    def __init__(self, context=None, namespace='default'):
        if context:
            config.load_kube_config(context=context)
        self.v1 = client.CoreV1Api()
        self.appsApi = client.AppsV1Api()
        self.batch_v1_beta = client.BatchV1beta1Api()
        self.logger = logging.getLogger(__name__)
        self.namespace = namespace

    def nodes(self):
        """ Lists all nodes found in the cluster """
        nodes = list()
        ret = self.v1.list_node()
        for item in ret.items:
            metadata = item.metadata
            status = item.status
            addresses = status.addresses
            addr_by_type = dict()
            for addr in addresses:
                addr_as_dict = addr.to_dict()
                addr_by_type[addr_as_dict["type"]] = addr_as_dict["address"]
            nodes.append(dict({
                "cluster_name": metadata.cluster_name or "",
                "name": metadata.name,
                "self_link": metadata.self_link,
                "uid": metadata.uid,
                "IP": addr_by_type["InternalIP"],
                "DNS": addr_by_type["InternalDNS"],
                "HOST": addr_by_type["Hostname"]
            }))
        return nodes

    def namespaces(self, pattern=".*"):
        """ returns all namespaces found in the current cluster
            that match a given pattern (or all otherwise) """

        p = re.compile(pattern)
        namespaces = list()

        ret = self.v1.list_namespace()
        for item in ret.items:
            metadata = item.metadata
            if not p.match(metadata.name):
                continue
            annotations = item.metadata.annotations
            area = "not set"
            area_owner_email = "not set"
            if annotations is not None:
                area = annotations.get("area-name", "")
                area_owner_email = annotations.get("area-owner-email", "")
            namespaces.append(dict({
                "name": metadata.name,
                "area": area,
                "area_owner_email": area_owner_email
            }))
        return namespaces

    def is_deployment_healthy(self, deployment) -> bool:
        if (deployment['DESIRED_REPLICAS'] ==
            deployment['status']['available_replicas'] ==
            deployment['status']['replicas'] ):
            return True
        else:
            return False


    def deployments_for_namespace(self, namespace):
        """list all deployments for the namespace """

        deployments = dict()
        ret = self.appsApi.list_namespaced_deployment(
            namespace=namespace)
        for deployment in ret.items:
            deployments[deployment.metadata.name] = dict({
                "DESIRED_REPLICAS":
                int(deployment.metadata.annotations.get('DESIRED_REPLICAS',
                    deployment.status.replicas if deployment.status.replicas else 0)),
                "name": deployment.metadata.name,
                "namespace": deployment.metadata.namespace,
                "labels": deployment.spec.template.metadata.labels,
                "status": dict({
                    "available_replicas": deployment.status.available_replicas,
                    "ready_replicas": deployment.status.ready_replicas,
                    "replicas": deployment.status.replicas,
                    "updated_replicas": deployment.status.updated_replicas
                    })
            })
        return deployments

    def namespaces_on_node(self):
        """ returns all namespaces found on node,
            currently not implemented"""
        pass

    def pods_in_namespace(self, namespace):
        self.namespace = namespace
        pods = list()
        ret = self.v1.list_namespaced_pod(namespace=namespace)
        for i in ret.items:
            pods.append(dict({
                "ip": i.status.pod_ip,
                "namespace": i.metadata.namespace,
                "name": i.metadata.name}))
        return pods

    def node_profile(self):
        # read_node
        pass

    def pod_profile(self, name, namespace):
        """ returns the pod profile as text"""
        profile = self.v1.read_namespaced_pod(name=name, namespace=namespace)
        prof_as_str = profile.to_str()
        prof_as_html = str()

        for line in prof_as_str.split("\n"):
            prof_as_html = prof_as_html + "<div>" + line + "</div>" + "\n"

        return prof_as_html

    def access_review(self, namespace, verb, resource):
        """returns if a verb on a resource is allowed or not"""

        auth = client.AuthorizationV1Api()
        access_review = client.V1LocalSubjectAccessReview(spec={
            "resourceAttributes": {
                'namespace': namespace,
                'verb': verb,
                'resource': resource
            },
        })
        response = auth.create_self_subject_access_review(access_review)
        return response.status.allowed

    def get_deployments_by_name(self,
        name: str,
        label_selector: str = None,
        ns: str = "default"
        ) -> list:

        field_selector = "metadata.name={name}".format(name=name)

        if label_selector:
            ret = self.appsApi.list_namespaced_deployment(
                    ns, 
                    field_selector=field_selector,
                    label_selector=label_selector)
        else:
            ret = self.appsApi.list_namespaced_deployment(ns, field_selector=field_selector)
        return ret

class Cron:
    """Class to manage cronjobs"""

    def __init__(self, context, namespace='default'):
        if context:
            config.load_kube_config(context=context)
        self.client_obj = client.api_client.ApiClient()
        self.batch_v1_beta = client.BatchV1beta1Api()
        self.logger = logging.getLogger(__name__)
        self.namespace = namespace

    def create_cronjob(self, 
            filename, 
            schedule,
            image, 
            command,
            name: str=None,
            service_account=None
            ) -> str:
        """Creates a new cronjob, returns the new cronjob name """

        try:
            if (name == None):
                name = filename + '-' + str(uuid1())[:8]
            cronjob = {
                "apiVersion": "batch/v1beta1",
                "kind": "CronJob",
                "metadata": {
                    "name": name,
                    "namespace": self.namespace,
                },
                "spec": {
                    "concurrencyPolicy": "Forbid",
                    "schedule": schedule,
                    "jobTemplate": {
                        "spec": {
                            "backoffLimit": 0,
                            "template": {
                                "metadata": {
                                    "labels": {
                                        "app": filename
                                    },
                                    "namespace": self.namespace
                                },
                                "spec": {
                                    "containers": [
                                        {
                                            "name": filename,
                                            "image": image,
                                            "image_pull_policy":"Always",
                                            "command":command,
                                            "resources": {
                                                "limits": {"memory": "999Mi"}, 
                                                "requests": {"memory": "999Mi"}
                                            }
                                        }
                                    ],
                                    "restartPolicy": "Never",
                                    "serviceAccountName": service_account,
                                    "terminationGracePeriodSeconds": 30
                               }
                            }
                        }
                    }
                }
            }

            utils.create_from_dict(
                    self.client_obj,
                    cronjob,
                    verbose=True,
                    namespace=self.namespace
            )
            return name
        except utils.FailToCreateError as err:
            self.logger.error("Error creating resource {}: {}".format(name, err))
            return None


    def delete_all_cronjobs(self):
        """ Deletes all cronjobs """

        cronjobs = self.get_cronjobs()
        return [self.delete_cronjob(c['name']) for c in cronjobs]


    def delete_cronjob(self, cronjobname):
        """ Deletes the selected cronjob """

        try:
            self.batch_v1_beta.delete_namespaced_cron_job(
            namespace=self.namespace,
            name=cronjobname
            )
            return cronjobname
        except Exception as err:
            self.logger.error("Error deleting cronjob {}: {}".format(cronjobname, err))
            return None


    def get_cronjobs(self):
        """ Returns all cronjobs """

        cronjobs = self.batch_v1_beta.list_namespaced_cron_job(
            namespace=self.namespace).items
        return [{
            "name" : i.metadata.name, 
            "schedule" : i.spec.schedule, 
            "status": {
                "active": "yes" if i.status.active else "no", 
                "last_schedule_time":str(i.status.last_schedule_time)
                }
            } for i in cronjobs]


    def __getitem__(self, position):
        return self.get_cronjobs()[position]

    def __len__(self):
        return len(self.get_cronjobs())


class Pod:
    """Class to manage pods"""

    def __init__(self, context, namespace='default'):
        if context:
            config.load_kube_config(context=context)
        self.client_obj = client.api_client.ApiClient()
        self.v1 = client.CoreV1Api()
        self.batch_v1_beta = client.BatchV1beta1Api()
        self.logger = logging.getLogger(__name__)
        self.namespace = namespace

    def create_pod(self, 
            image, 
            command,
            name: str,
            service_account=None
            ) -> str:
        try:
            id = name + '-' + str(uuid1())[:8]
            pod = {
                "apiVersion": "v1",
                "kind": "Pod",
                "metadata": {
                    "name": id,
                    "namespace": self.namespace,
                },
                "spec": {
                    "containers": [
                        {
                            "name": id,
                            "image": image,
                            "command":command,
                            "resources": {
                                "limits": {"memory": "999Mi"}, 
                                "requests": {"memory": "999Mi"}
                            }
                        }
                    ],
                    "restartPolicy": "Never",
                    "terminationGracePeriodSeconds": 30
                }
            }

            utils.create_from_dict(
                    self.client_obj,
                    pod,
                    verbose=True,
                    namespace=self.namespace
            )
            return id
        except utils.FailToCreateError as err:
            self.logger.error("Error creating resource {}: {}".format(name, err))
            return None


